module("L_Powermax", package.seeall)

-------------------------------------
-- PowerMax control via Mios Vera  --
-- Based on original code from utz --
-------------------------------------

local bitw = require("bit")

local PLUGIN_VERSION = "1.0 beta"

local Queue = {}
function Queue.new()
	return {first = 0, last = -1}
end

function Queue.pushright(list, value)
	local last = list.last + 1
	list.last = last
	list[last] = value
end
    
function Queue.popleft(list)
   local first = list.first
   if first > list.last then return nil end
   local value = list[first]
   list[first] = nil        -- to allow garbage collection
   list.first = first + 1
   return value
end

function Queue.len(list)
   return list.last - list.first
end

-- DEVICES --
local pmPanelDev = 0
local pmPartitionDev_t  = { 0, 0, 0 }
local pmSensorDev_t = {} -- id, name, stype, sid, sname, ztype, zname, zchime, partition
local pmSirenDev_t = {}
local pmPanelTypeNr = 3 -- assume Powerlink Pro (no partitions) by default
-- MESSAGE HANDLING --
local pmIncomingPdu = ""
local pmIncomingPduLen = nil
local pmLastPDU = ""
local pmOutgoingQueue = Queue.new()
local pmWaitingForResponse = 0
local pmLastSendMsg = ""
local pmExpectedResponse = ""
local pmSendMsgRetries = 0
local pmCrcErrorCount = 0
local TIMEOUT = 9 -- wait max. 9 sec for a response (actually 20 due to timer interval)
local MSG_RETRIES = 4
local MAX_CRC_ERROR = 5
local MAX_EXCEPTIONS = 3
-- SETTINGS --
local pmSettings_t = {}
local pmForcedDisarmCode = string.char(0x00, 0x00)
local pmPincode_t = { string.char(0x00, 0x00) }
local pmBellTime = nil;
local pmSilentPanic = false;
local pmQuickArm = false;
local pmBypassOff = false;
local pmLang = "EN"
local pmMotionOffDelay = 3
local pmRemoteArm = "false"
local pmSensorArm = "auto"
local pmSensorShowBypass = true
-- EVENT LOG --
local pmEventLog = ""
local pmEventCnt
-- POWERLINK --
local pmDownloadCode = "VP" -- Vera Powerlink
local pmPowerlinkMode = false
local pmLastKeepAlive = nil
local pmSirenActive = nil
local pmPowerlinkRetry = 0
local PL_RETRIES = 3
--- DEBUGGING ---
local pmLogDebug = true
local pmFilenameLog = "/var/log/cmh/powermax_pdu.txt"
local pmFilenameSettings = "/var/log/cmh/powermax_settings.txt"

local pmPanelInit_t = {
	{ "PluginVersion", PLUGIN_VERSION, false }, -- false: update
   { "MotionOffDelay", pmMotionOffDelay, true }, -- true: create only
   { "PluginLanguage", pmLang, true },
	{ "PluginDebug", "0", true },
   { "ForceStandard", "0", true },
   { "AutoCreate", "1", true },
   { "AutoSyncTime", "1", true },
   { "EnableRemoteArm", pmRemoteArm, true },
   { "SensorArm", pmSensorArm, true },
   { "DoorZones", "", true },
   { "MotionZones", "", true },
   { "SmokeZones", "", true },
   { "Devices", "", true },
   { "PowerlinkMode", "Unknown", false },
   { "PanelStatusCode", "", true },
   { "PanelStatusData", "", true },
   { "PanelAlarmType", "", true },
   { "PanelTroubleType", "", true },
   { "EventLog", "", true }
}

local PANEL_SID        = "urn:micasaverde-com:serviceId:PowermaxAlarmPanel1"
local PARTITION_SID    = "urn:micasaverde-com:serviceId:AlarmPartition2"
local SECURITY_SID     = "urn:micasaverde-com:serviceId:SecuritySensor1"
local SENSOR_SID       = "urn:micasaverde-com:serviceId:HaDevice1"
local KEYPAD_SID       = "urn:micasaverde-com:serviceId:Keypad1"
     
local X10DEVICE_SID    = "urn:upnp-org:serviceId:SwitchPower1"
local X10DEVICEDIM_SID = "urn:upnp-org:serviceId:Dimming1"
local ENERGY_SID       = "urn:micasaverde-com:serviceId:EnergyMetering1"
local SIREN_SID        = "urn:upnp-org:serviceId:SwitchPower1"

local CON_X10SW_INIT  = X10DEVICE_SID .. ",Target=0\n" .. X10DEVICE_SID .. ",Status=0\n" .. ENERGY_SID .. ",Watts=\n" .. ENERGY_SID .. ",UserSuppliedWattage="
local CON_X10DIM_INIT = CON_X10SW_INIT .. "\n" .. X10DEVICEDIM_SID .. ",LoadLevelTarget=0\n" .. X10DEVICEDIM_SID .. ",LoadLevelStatus=0"
local CON_SENSOR_INIT = SECURITY_SID .. ",Tripped=0"
local CON_BATSENSOR_INIT = CON_SENSOR_INIT .. "\n" .. SENSOR_SID .. ",BatteryLevel=100"
local CON_SIREN       = X10DEVICE_SID .. ",Target=0\n" .. X10DEVICE_SID .. ",Status=0\n"
local CON_KEYPAD      = KEYPAD_SID .. ",Status=0"

local pmDevices_t = {
   VAR_PARTITION = { "urn:schemas-micasaverde-com:device:AlarmPartition:2", "D_PowermaxPartition2.xml", "", "", true },
   VAR_SWITCH    = { "urn:schemas-micasaverde-com:device:BinaryLight:1", "D_BinaryLight1.xml", "", CON_X10SW_INIT, false },
   VAR_DIM       = { "urn:schemas-micasaverde-com:device:DimmableLight:1", "D_DimmableLight1.xml", "",CON_X10DIM_INIT, false },
   VAR_MAGNET    = { "urn:schemas-micasaverde-com:device:DoorSensor:1", "D_DoorSensor1.xml", "", CON_BATSENSOR_INIT, false },
   VAR_MOTION    = { "urn:schemas-micasaverde-com:device:MotionSensor:1", "D_MotionSensor1.xml", "", CON_BATSENSOR_INIT, false },
   VAR_SMOKE     = { "urn:schemas-micasaverde-com:device:SmokeSensor:1", "D_SmokeSensor1.xml", "", CON_BATSENSOR_INIT, false },
   VAR_GAS       = { "urn:schemas-micasaverde-com:device:SmokeSensor:1", "D_SmokeSensor1.xml", "", CON_BATSENSOR_INIT, false },
   VAR_SIREN     = { "urn:schemas-micasaverde-com:device:Siren:1", "D_Siren1.xml", "", CON_SIREN, false },
   VAR_KEYPAD    = { "urn:schemas-micasaverde-com:service:Keypad:1", "D_Keypad1.xml", "", CON_KEYPAD, false }
}

-- PowerMax messages we send
-- E1: message string
-- E2: expected response after ACK (nil if none)
local pmSendMsg_t = {
   MSG_INIT      = { string.char(0xAB, 0x0A, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x43), nil },
   MSG_RESTORE   = { string.char(0xAB, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x43), 0xA5 },
   MSG_ENROLL    = { string.char(0xAB, 0x0A, 0x00, 0x00) .. pmDownloadCode .. string.char(0x00, 0x00, 0x00, 0x00, 0x00, 0x43), nil },
   MSG_EVENTLOG  = { string.char(0xA0, 0x00, 0x00, 0x00) .. "pin" .. string.char(0x00, 0x00, 0x00, 0x00, 0x00, 0x43), 0xA0 },
   MSG_ARM       = { string.char(0xA1, 0x00, 0x00) .. "armpin" .. string.char(0x00, 0x00, 0x00, 0x00, 0x00, 0x43), nil },
   MSG_STATUS    = { string.char(0xA2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x43), 0xA5 },
   MSG_X10PGM    = { string.char(0xA4, 0x00, 0x00, 0x00, 0x00, 0x00) .. "cmddevice" .. string.char(0x00, 0x00, 0x43), nil },
   MSG_BYPASSEN  = { string.char(0xAA) .. "pinbypass" .. string.char(0x00, 0x00, 0x00, 0x00, 0x43), nil },
   MSG_BYPASSDIS = { string.char(0xAA) .. "pin" .. string.char(0x00, 0x00, 0x00, 0x00) .. "bypass" .. string.char(0x43), nil },
   MSG_DOWNLOAD  = { string.char(0x24, 0x00, 0x00) .. pmDownloadCode.. string.char(0x00, 0x00, 0x00, 0x00, 0x00, 0x00), 0x3C },
   MSG_SETTIME   = { string.char(0x46, 0xF8, 0x00) .. "time" .. string.char(0xFF, 0xFF), nil },
   MSG_DL        = { string.char(0x3E) .. "item" .. string.char(0xB0, 0x00, 0x00, 0x00, 0x00, 0x00), 0x3F},
   MSG_SER_TYPE  = { string.char(0x5A, 0x30, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00), 0x33},
   MSG_START     = { string.char(0x0A), 0x33 },
   MSG_EXIT      = { string.char(0x0F), nil }
}

pmDownloadItem_t = {
   MSG_DL_COMMDEF    = string.char(0x01, 0x01, 0x1E, 0x00),
   MSG_DL_PHONENRS   = string.char(0x36, 0x01, 0x20, 0x00),
   MSG_DL_PINCODES   = string.char(0xFA, 0x01, 0x10, 0x00),
   MSG_DL_PGMX10     = string.char(0x14, 0x02, 0xD5, 0x00),
   MSG_DL_PANELFW    = string.char(0x00, 0x04, 0x20, 0x00),
   MSG_DL_SERIAL     = string.char(0x30, 0x04, 0x08, 0x00),
   MSG_DL_ZONES      = string.char(0x00, 0x09, 0x78, 0x00),
   MSG_DL_KEYFOBS    = string.char(0x78, 0x09, 0x40, 0x00),
   MSG_DL_2WKEYPAD   = string.char(0x00, 0x0A, 0x08, 0x00),
   MSG_DL_1WKEYPAD   = string.char(0x20, 0x0A, 0x40, 0x00),
   MSG_DL_SIRENS     = string.char(0x60, 0x0A, 0x08, 0x00),
   MSG_DL_X10NAMES   = string.char(0x30, 0x0B, 0x10, 0x00),
   MSG_DL_ZONENAMES  = string.char(0x40, 0x0B, 0x1E, 0x00),
   MSG_DL_EVENTLOG   = string.char(0xDF, 0x04, 0x28, 0x03),
   MSG_DL_ZONESTR    = string.char(0x00, 0x19, 0x00, 0x02),
   MSL_DL_ZONECUSTOM = string.char(0xA0, 0x1A, 0x50, 0x00),
   MSG_DL_ALL        = string.char(0x00, 0x00, 0x00, 0xFF)
}

-- Message types we can receive with their length (nil=unknown), handlers and whether they need an ACK
local pmReceiveMsg_t = {
   [0x02] = { nil, pmHandleAck, false },
   [0x06] = { nil, pmHandleTimeout, false },
   [0x08] = { nil, pmHandleDenied, true },
   [0x0B] = { nil, pmHandleStop, true },
   [0x33] = { 14, pmHandleSettings, true },
   [0x3C] = { 14, pmHandleInfo, true },
   [0x3F] = { nil, pmHandleSettingsItem, true },
   [0xA0] = { 15, pmHandleEventlog, true },
   [0xA5] = { 15, pmHandleStatus, true },
   [0xA7] = { 15, pmHandlePanel, true },
   [0xAB] = { 15, pmHandlePowerlink, false }
}

local pmLogEvent_t = { 
   EN = {
   "None", "Interior Alarm", "Perimeter Alarm", "Delay Alarm", "24h Silent Alarm", "24h Audible Alarm",
   "Tamper", "Control Panel Tamper", "Tamper Alarm", "Tamper Alarm", "Communication Loss", "Panic From Keyfob",
   "Panic From Control Panel", "Duress", "Confirm Alarm", "General Trouble", "General Trouble Restore", 
   "Interior Restore", "Perimeter Restore", "Delay Restore", "24h Silent Restore", "24h Audible Restore",
   "Tamper Restore", "Control Panel Tamper Restore", "Tamper Restore", "Tamper Restore", "Communication Restore",
   "Cancel Alarm", "General Restore", "Trouble Restore", "Not used", "Recent Close", "Fire", "Fire Restore", 
   "No Active", "Emergency", "No used", "Disarm Latchkey", "Panic Restore", "Supervision (Inactive)",
   "Supervision Restore (Active)", "Low Battery", "Low Battery Restore", "AC Fail", "AC Restore",
   "Control Panel Low Battery", "Control Panel Low Battery Restore", "RF Jamming", "RF Jamming Restore",
   "Communications Failure", "Communications Restore", "Telephone Line Failure", "Telephone Line Restore",
   "Auto Test", "Fuse Failure", "Fuse Restore", "Keyfob Low Battery", "Keyfob Low Battery Restore", "Engineer Reset",
   "Battery Disconnect", "1-Way Keypad Low Battery", "1-Way Keypad Low Battery Restore", "1-Way Keypad Inactive",
   "1-Way Keypad Restore Active", "Low Battery", "Clean Me", "Fire Trouble", "Low Battery", "Battery Restore",
   "AC Fail", "AC Restore", "Supervision (Inactive)", "Supervision Restore (Active)", "Gas Alert", "Gas Alert Restore",
   "Gas Trouble", "Gas Trouble Restore", "Flood Alert", "Flood Alert Restore", "X-10 Trouble", "X-10 Trouble Restore",
   "Arm Home", "Arm Away", "Quick Arm Home", "Quick Arm Away", "Disarm", "Fail To Auto-Arm", "Enter To Test Mode",
   "Exit From Test Mode", "Force Arm", "Auto Arm", "Instant Arm", "Bypass", "Fail To Arm", "Door Open",
   "Communication Established By Control Panel", "System Reset", "Installer Programming", "Wrong Password",
   "Not Sys Event", "Not Sys Event", "Extreme Hot Alert", "Extreme Hot Alert Restore", "Freeze Alert", 
   "Freeze Alert Restore", "Human Cold Alert", "Human Cold Alert Restore", "Human Hot Alert",
   "Human Hot Alert Restore", "Temperature Sensor Trouble", "Temperature Sensor Trouble Restore"
}, NL = {
   "Geen", "In alarm", "In alarm", "In alarm", "In alarm", "In alarm", 
   "Sabotage alarm", "Systeem sabotage", "Sabotage alarm", "Add user", "Communicate fout", "Paniekalarm",
   "Code bedieningspaneel paniek", "Dwang", "Bevestig alarm", "Successful U/L", "Probleem herstel",
   "Herstel", "Herstel", "Herstel", "Herstel", "Herstel", 
   "Sabotage herstel", "Systeem sabotage herstel", "Sabotage herstel", "Sabotage herstel", "Communicatie herstel",
   "Stop alarm", "Algemeen herstel", "Brand probleem herstel", "Systeem inactief", "Recent close", "Brand", "Brand herstel",
   "Niet actief", "Noodoproep", "Remove user", "Controleer code", "Bevestig alarm", "Supervisie", 
   "Supervisie herstel", "Batterij laag", "Batterij OK", "230VAC uitval", "230VAC herstel",
   "Controlepaneel batterij laag", "Controlepaneel batterij OK", "Radio jamming", "Radio herstel",
   "Communicatie mislukt", "Communicatie hersteld", "Telefoonlijn fout", "Telefoonlijn herstel", 
   "Automatische test", "Zekeringsfout", "Zekering herstel", "Batterij laag", "Batterij OK", "Monteur reset",
   "Accu vermist", "Batterij laag", "Batterij OK", "Supervisie", 
   "Supervisie herstel", "Lage batterij bevestiging", "Reinigen", "Probleem", "Batterij laag", "Batterij OK",
   "230VAC uitval", "230VAC herstel", "Supervisie", "Supervisie herstel", "Gas alarm", "Gas herstel",
   "Gas probleem", "Gas probleem herstel", "Lekkage alarm", "Lekkage herstel", "Probleem", "Probleem herstel", 
   "Deelschakeling", "Ingeschakeld", "Snel deelschakeling", "Snel ingeschakeld", "Uitgezet", "Inschakelfout (auto)", "Test gestart",
   "Test gestopt", "Force aan", "Geheel in (auto)", "Onmiddelijk", "Overbruggen", "Inschakelfout",
   "Log verzenden", "Systeem reset", "Installateur programmeert", "Foutieve code", "Overbruggen"
}}

local pmLogUser_t = {
   "System ", "Zone 01", "Zone 02", "Zone 03", "Zone 04", "Zone 05", "Zone 06", "Zone 07", "Zone 08",
   "Zone 09", "Zone 10", "Zone 11", "Zone 12", "Zone 13", "Zone 14", "Zone 15", "Zone 16", "Zone 17", "Zone 18",
   "Zone 19", "Zone 20", "Zone 21", "Zone 22", "Zone 23", "Zone 24", "Zone 25", "Zone 26", "Zone 27", "Zone 28", 
   "Zone 29", "Zone 30", "Fob  01", "Fob  02", "Fob  03", "Fob  04", "Fob  05", "Fob  06", "Fob  07", "Fob  08", 
   "User 01", "User 02", "User 03", "User 04", "User 05", "User 06", "User 07", "User 08", "Pad  01", "Pad  02",
   "Pad  03", "Pad  04", "Pad  05", "Pad  06", "Pad  07", "Pad  08", "Sir  01", "Sir  02", "2Pad 01", "2Pad 02",
   "2Pad 03", "2Pad 04", "X10  01", "X10  02", "X10  03", "X10  04", "X10  05", "X10  06", "X10  07", "X10  08",
   "X10  09", "X10  10", "X10  11", "X10  12", "X10  13", "X10  14", "X10  15", "PGM    ", "GSM    ", "P-LINK ",
   "PTag 01", "PTag 02", "PTag 03", "PTag 04", "PTag 05", "PTag 06", "PTag 07", "PTag 08" 
}

local pmSysStatus_t = { EN = {
   "Disarmed", "Home Exit Delay", "Away Exit Delay", "Entry Delay", "Armed Home", "Armed Away", "User Test",
   "Downloading", "Programming", "Installer", "Home Bypass", "Away Bypass", "Ready", "Not Ready", "??", "??", 
   "Disarmed Instant", "Home Instant Exit Delay", "Away Instant Exit Delay", "Entry Delay Instant", "Armed Home Instant", 
   "Armed Away Instant"
}, NL = {
   "Uitgeschakeld", "Deel uitloopvertraging", "Totaal uitloopvertraging", "Inloopvertraging", "Deel ingeschakeld",
   "Totaal ingeschakeld", "Gebruiker test", "Downloaden", "Programmeren", "Monteurmode", "Deel met overbrugging",
   "Totaal met overbrugging", "Klaar", "Niet klaar", "??", "??", "Direct uitschakelen", "Direct Deel uitloopvertraging",
   "Direct Totaal uitloopvertraging", "Direct inloopvertraging", "Direct Deel", "Direct Totaal"
}}

local pmSysStatusFlags_t = { 
   EN = {
   "Ready", "Alert in memory", "Trouble", "Bypass on", "Last 10 seconds", "Zone event", "Status changed", "Alarm event"
}, NL = {
   "Klaar", "Alarm in geheugen", "Probleem", "Overbruggen aan", "Laatste 10 seconden", "Zone verstoord", "Status gewijzigd",
   "Alarm actief"
}}

local pmArmed_t = {
   [0x03] = "", [0x04] = "", [0x05] = "", [0x0A] = "", [0x0B] = "", [0x13] = "", [0x14] = "", [0x15] = ""
}

local pmArmMode_t = {
   Disarmed = 0x00, Stay = 0x04, Armed = 0x05, UserTest = 0x06, StayInstant = 0x14, ArmedInstant = 0x15,
   Night = 0x04, NightInstant = 0x14
}

local pmDetailedArmMode_t = {
   "Disarmed", "ExitDelay", "ExitDelay", "EntryDelay", "Stay", "Armed", "UserTest", "NotReady", "NotReady", "NotReady",
   "Force", "Force", "Ready", "NotReady", "??", "??", "Disarmed", "ExitDelay", "ExitDelay", "EntryDelay", "StayInstant",
   "ArmedInstant"
} -- Not used: Night, NightInstant, Vacation

local pmEventType_t = { 
   EN = {
   "None", "Tamper Alarm", "Tamper Restore", "Open", "Closed", "Violated (Motion)", "Panic Alarm", "RF Jamming",
   "Tamper Open", "Communication Failure", "Line Failure", "Fuse", "Not Active", "Low Battery", "AC Failure", 
   "Fire Alarm", "Emergency", "Siren Tamper", "Siren Tamper Restore", "Siren Low Battery", "Siren AC Fail"
}, NL = {
   "Geen", "Sabotage alarm", "Sabotage herstel", "Open", "Gesloten", "Verstoord (beweging)", "Paniek alarm", "RF verstoring",
   "Sabotage open", "Communicatie probleem", "Lijnfout", "Zekering", "Niet actief", "Lage batterij", "AC probleem",
   "Brandalarm", "Noodoproep", "Sirene sabotage", "Sirene sabotage herstel", "Sirene lage batterij", "Sirene AC probleem"
}}

local pmPanelAlarmType_t = {
   [0x01] = "Intruder", [0x02] = "Intruder", [0x03] = "Intruder", [0x04] = "Intruder", [0x05] = "Intruder", [0x06] = "Tamper", 
   [0x07] = "Tamper", [0x08] = "Tamper", [0x09] = "Tamper", [0x0B] = "Panic", [0x0C] = "Panic", [0x20] = "Fire", 
   [0x23] = "Emergency", [0x49] = "Gas", [0x4D] = "Flood"
}

local pmPanelTroubleType_t = {
   [0x0A] = "Communication", [0x0F] = "General", [0x29] = "Battery", [0x2B] = "Power", [0x2D] = "Battery", [0x2F] = "Jamming", 
   [0x31] = "Communication", [0x33] = "Telephone", [0x36] = "Power", [0x38] = "Battery", [0x3B] = "Battery", [0x3C] = "Battery",
   [0x40] = "Battery", [0x43] = "Battery"
}

local pmPanelType_t = {
   [1] = "PowerMax", [2] = "PowerMax+", [3] = "PowerMax Pro", [4] = "PowerMax Complete", [5] = "PowerMax Pro Part",
   [6] = "PowerMax Complete Part", [7] = "PowerMax Express", [8] = "PowerMaster10", [9] = "PowerMaster30"
}

-- Config for each panel type (1-9)
local pmPanelConfig_t = {
   CFG_PARTITIONS = {   0,   0,   0,   0,   3,   3,   0,   3,   3 },
   CFG_EVENTS     = { 250, 250, 250, 250, 250, 250, 250, 250,1000 },
   CFG_KEYFOBS    = {   8,   8,   8,   8,   8,   8,   8,   8,  32 },
   CFG_1WKEYPADS  = {   8,   8,   8,   8,   8,   8,   8,   0,   0 },
   CFG_2WKEYPADS  = {   2,   2,   2,   2,   2,   2,   2,   8,  32 },
   CFG_SIRENS     = {   2,   2,   2,   2,   2,   2,   2,   4,   8 },
   CFG_USERCODES  = {   8,   8,   8,   8,   8,   8,   8,   8,  48 },
   CFG_PROXTAGS   = {   0,   0,   8,   0,   8,   8,   0,   8,  32 },
   CFG_WIRELESS   = {  28,  28,  28,  28,  28,  28,  28,  30,  64 },
   CFG_WIRED      = {   2,   2,   2,   2,   2,   2,   1,   1,   2 },
   CFG_ZONECUSTOM = {   0,   5,   5,   5,   5,   5,   5,   5,   5 }
}

local pmPanelName_t = {
   ["0000"] = "PowerMax", ["0001"] = "PowerMax LT", ["0004"] = "PowerMax A", ["0005"] = "PowerMax 108", ["0006"] = "PowerMax LT 108",
   ["0009"] = "PowerMax B", ["000a"] = "PowerMax A 112", ["000b"] = "PowerMax 112", ["000c"] = "PowerMax LT 112", 
   ["000f"] = "PowerMax B 112", ["0014"] = "PowerMax A 113", ["0015"] = "PowerMax 113-4", ["0016"] = "PowerMax 113-6",
   ["0017"] = "PowerArt 113", ["0018"] = "PowerMax SC 113", ["0019"] = "PowerMax SK 113", ["001a"] = "PowerMax 113-4-SV",
   ["001b"] = "PowerMax 113-4-T", ["001e"] = "PowerMax WSS", ["001f"] = "PowerMax SmithAndWesson", ["0100"] = "PowerMax+",
   ["0103"] = "PowerMax+ UK", ["0104"] = "PowerMax+ JP", ["0106"] = "PowerMax+ CTA", ["0108"] = "PowerMax+ GSM", ["010a"] = "PowerMax+ SH",
   ["010b"] = "PowerMax+ CF", ["0112"] = "PowerMax+ WSS", ["0113"] = "PowerMax+ 2INST", ["0114"] = "PowerMax+ HL",
   ["0115"] = "PowerMax+ UK2", ["0116"] = "PowerMax+ 2INST3", ["0118"] = "PowerMax+ CF2", ["0119"] = "PowerMax+ 2INST2",
   ["011a"] = "PowerMax+ GSM2", ["011c"] = "PowerMax+ WSS33", ["011d"] = "PowerMax+ UK3", ["0120"] = "PowerMax+ 2INST33/ALPH",
   ["0121"] = "PowerMax+ GSM3", ["0122"] = "PowerMax+ CF3", ["0124"] = "PowerMax+ UK33", ["0127"] = "PowerMax+ 2INST_MONITOR39",
   ["0128"] = "PowerMax+ KeyOnOff", ["0129"] = "PowerMax+ 2INST_MONITOR41", ["012a"] = "PowerMax+ 2INST_MONITOR42", 
   ["012b"] = "PowerMax+ 2INST_3_4_17", ["012c"] = "PowerMax+ 1_44_0", ["012d"] = "PowerMax+ CF3_1_45_0", ["012e"] = "PowerMax+ SA_1_46",
   ["012f"] = "PowerMax+ UK_1_47/UK_3_4_17", ["0130"] = "PowerMax+ SA UK_1_48", ["0132"] = "PowerMax+ KeyOnOff_1_50",
   ["0201"] = "PowerMax Pro", ["0202"] = "PowerMax Pro-Nuon", ["0204"] = "PowerMax Pro-PortugalTelecom",
   ["020a"] = "PowerMax Pro-PortugalTelecom2", ["020c"] = "PowerMax HW-V9 Pro", ["020d"] = "PowerMax ProSMS",
   ["0214"] = "PowerMax Pro-PortugalTelecom_4_5_02", ["0216"] = "PowerMax HW-V9_4_5_02 Pro", ["0217"] = "PowerMax ProSMS_4_5_02",
   ["0218"] = "PowerMax UK_DD243_4_5_02 Pro", ["021b"] = "PowerMax Pro-Part2 M 2_27", ["0223"] = "PowerMax Pro Bell-Canada",
   ["0301"] = "PowerMax LE", ["0302"] = "PowerMax LE_NV", ["0303"] = "PowerMax LE_PortugalTelecom", ["0307"] = "PowerMax LE_1_0_07",
   ["0308"] = "PowerMax LE_NV_1_0_07", ["030a"] = "PowerMax LE_UK_DD243_1_1_03", ["030b"] = "PowerMax LE_COUNTERFORCE_1_0_06",
   ["0401"] = "PowerMax Pro-Part", ["0402"] = "PowerMax Pro-Part CellAdaptor", ["0405"] = "PowerMax Pro-Part_5_0_8",
   ["0406"] = "PowerMax Pro-Part CellAdaptor_5_2_04", ["0407"] = "PowerMax KeyOnOff Pro-Part_5_0_8",
   ["0408"] = "PowerMax UK Pro-Part_5_0_8", ["0409"] = "PowerMax UK-Sector Pro-Part_5_0_8", ["040a"] = "PowerMax CP1 4_10",
   ["040c"] = "PowerMax Pro-Part CellKey_4_12", ["040d"] = "PowerMax UK 4_13", ["040e"] = "PowerMax UK-Sector 4_14",
   ["040f"] = "PowerMax UK 4_15", ["0410"] = "PowerMax CP1 4_16", ["0411"] = "PowerMax Pro-Part NUON Key 4_17",
   ["0433"] = "PowerMax Pro-Part2 4_51", ["0434"] = "PowerMax UK Pro-Part2 4_52", ["0436"] = "PowerMax Pro-Part2 4_54",
   ["0437"] = "PowerMax Pro-Part2 4_55", ["0438"] = "PowerMax Pro-Part2 4_56", ["0439"] = "PowerMax Pro-Part2 4_57", 
   ["043a"] = "PowerMax 4_58", ["043c"] = "PowerMax 4_60", ["043e"] = "PowerMax Pro-Part2 4_62", ["0440"] = "PowerMax Pro-Part2 4_64",
   ["0442"] = "PowerMax 4_66", ["0443"] = "PowerMax 4_67", ["0444"] = "PowerMax 4_68", ["0445"] = "PowerMax 4_69",
   ["0446"] = "PowerMax Pro-Part2 4_70", ["0447"] = "PowerMax 4_71", ["0449"] = "PowerMax 4_73", ["044b"] = "PowerMax 4_75",
   ["0451"] = "PowerMax 4_81", ["0452"] = "PowerMax 4_82", ["0503"] = "PowerMax UK LE-Part 1_5_00", ["050a"] = "PowerMax LE-Part GPRS",
   ["050b"] = "PowerMax LE-Part NV GPRS", ["050c"] = "PowerMax LE-Part GPRS-", ["050d"] = "PowerMax LE-Part NV GPRS-",
   ["050e"] = "PowerMax LE-Part GPRS- UK", ["0511"] = "PowerMax LE-Part GPRS 5_17", ["0512"] = "PowerMax LE-Part GPRS- UK_5_18",
   ["0533"] = "PowerMax Complete 5_51", ["0534"] = "PowerMax LE-Part2 5_52", ["0536"] = "PowerMax Complete 5_54",
   ["0537"] = "PowerMax LE-Part2 5_55", ["053a"] = "PowerMax Complete 5_58", ["053b"] = "PowerMax Complete 5_59",
   ["053e"] = "PowerMax Complete 5_62", ["053f"] = "PowerMax LE-Part2 5_63_2", ["0540"] = "PowerMax LE-Part2 5_64",
   ["0543"] = "PowerMax Complete 5_67", ["0544"] = "PowerMax LE-Part2 5_68", ["0545"] = "PowerMax LE-Part2 5_69",
   ["0546"] = "PowerMax Complete 5_70", ["0547"] = "PowerMax Complete 5_71", ["0549"] = "PowerMax Complete 5_73",
   ["054b"] = "PowerMax Complete 5_75", ["0601"] = "PowerMax Express", ["0603"] = "PowerMax Express_6_3",
   ["0605"] = "PowerMax Express OEM_6_5", ["0607"] = "PowerMax Express BBA_6_7", ["0608"] = "PowerMax Express 6_8",
   ["0609"] = "PowerMax Express OEM_6_9", ["060b"] = "PowerMax Express BBA_6_11", ["0633"] = "PowerMax Express Part2 6_51",
   ["063b"] = "PowerMax Express Part2 6_59", ["063e"] = "PowerMax Express Part2 6_62", ["0645"] = "PowerMax Express 6_69",
   ["0645"] = "PowerMax Express Part2 6_69", ["0647"] = "PowerMax Express Part2 6_71", ["0648"] = "PowerMax Express Part2 6_72",
   ["0649"] = "PowerMax Express Part2 6_73", ["064a"] = "PowerMax Secure Part2 6_74", ["064c"] = "PowerMax Express Part2 6_76",
   ["064f"] = "PowerMax Express Part2 6_79", ["0651"] = "PowerMax Express Part2 6_81", ["0652"] = "PowerMax Express Part2 6_82",
   ["0654"] = "PowerMax Express 6_84", ["0655"] = "PowerMax Express 6_85", ["0701"] = "PowerMax PowerCodeG_7_1",
   ["0702"] = "PowerMax PowerCodeG_7_2", ["0704"] = "PowerMaster10 7_4", ["070c"] = "PowerMaster10 7_12",
   ["070f"] = "PowerMaster10 7_15", ["0710"] = "PowerMaster10 7_16", ["0800"] = "PowerMax Pro-Part2 2_27",
   ["0802"] = "PowerMax PowerCodeG_8_2", ["0803"] = "PowerMaster30 8_3", ["080f"] = "PowerMaster30 8_15",
   ["0810"] = "PowerMaster30 8_16",
}

local pmZoneType_t = { 
   EN = {
   "Non-Alarm", "Emergency", "Flood", "Gas", "Delay 1", "Delay 2", "Interior-Follow", "Perimeter",
   "Perimeter-Follow", "24 Hours Silent", "24 Hours Audible", "Fire", "Interior"
}, NL = {
   "Geen alarm", "Noodtoestand", "Water", "Gas", "Vertraagd 1", "Vertraagd 2", "Interieur volg", "Omtrek",
   "Omtrek volg", "24 uurs stil", "24 uurs luid", "Brand", "Interieur"
}}

-- Zone names are taken from the panel, so no langauage support needed
local pmZoneName_t = {
   "Attic", "Back door", "Basement", "Bathroom", "Bedroom", "Child room", "Closet", "Den", "Dining room", "Downstairs", 
   "Emergency", "Fire", "Front door", "Garage", "Garage door", "Guest room", "Hall", "Kitchen", "Laundry room", "Living room", 
   "Master bathroom", "Master bedroom", "Office", "Upstairs", "Utility room", "Yard", "Custom 1", "Custom 2", "Custom 3",
   "Custom4", "Custom 5", "Not Installed"
}

local pmZoneChime_t = {
   "Chime Off", "Melody Chime", "Zone Name Chime"
}

-- Note: names need to match to VAR_xxx
local pmZoneSensor_t = {
   [0x3] = "Motion", [0x4] = "Motion", [0x5] = "Magnet", [0x6] = "Magnet", [0xA] = "Smoke", [0xB] = "Gas", [0xF] = "Siren"
} -- unknown to date: Control Panel, Push Button, Flood, Universal

-- Update a system variable only if the value will change
function updateIfNeeded(sid, var, newVal, id, createOnly)
	local curVal = luup.variable_get(sid, var, id)
   local valUpdate = (curVal == nil) or ((createOnly ~= true) and (curVal ~= tostring(newVal)) or false)
	if (valUpdate == true) then
		luup.variable_set(sid, var, newVal, id)
		return true
	end
	return false
end

-- Find child device (Taken from GE Caddx Panel but comes originally from guessed)
function findChild(deviceId, label)
	for k, v in pairs(luup.devices) do
	   if (v.device_num_parent == deviceId and v.id == label) then
		   return k
	   end
	end
end

-- fileExists: check if a device file is available
function fileExists(name, path)
   local s = name .. ".lzo"
   if (path == nil) then
      s = "/etc/cmh-lu/" .. s
   else
      s = path .. s
   end
   local f = io.open(s, "r")
   if (f ~= nil) then
      io.close(f)
      return true
   else
      if (string.find(s, "ludl") == nil) then
         return fileExists(name, "/etc/cmh-ludl/")
      else
         return false
      end      
   end
end

-- Control debug to /var/log/cmh/LuaUPnP.log (taken from GE Caddx Panel)
function debug(s)
	if (pmLogDebug) then
	   luup.log("POWERMAX: " .. s)
	end
end

-- pmDumpSettings
function pmDumpSettings()
   if (pmLogDebug) then
      local dumpfile = pmFilenameSettings
      local outf = io.open(dumpfile , 'w')
      debug("Dumping PowerMax settings to file")
      outf:write(string.format("PowerMax settings on %s\n", os.date('%Y-%m-%d %H:%M:%S')))
      for i = 0, 0xFF do
         if (pmSettings_t[i] ~= nil) then
            for j = 0, 0x0F do
               local s = ""
               outf:write(string.format("%08X: ", i * 0x100 + j * 0x10))
               for k = 1, 0x10 do
                  local byte = string.byte(pmSettings_t[i], j * 0x10 + k)
                  outf:write(string.format(" %02X", byte))
                  s = (byte < 0x20 or (byte >= 0x80 and byte < 0xA0)) and (s .. ".") or (s .. string.char(byte))
               end
               outf:write("  " .. s .. "\n")
            end
         end
      end
      outf:close()
   end
end

-- PDUs can be logged in a file. We use this to discover new never before seen
-- PowerMax messages we need to decode and make sense of in long evenings...
function pmLogPdu(PDU, direction)
   if (pmLogDebug) then
      local logfile = pmFilenameLog
      -- empty file if it reaches 250kb
      local outf = io.open(logfile , 'a')
      local filesize = outf:seek("end")
      outf:close()
      if (filesize > 250000) then
         local outf = io.open(logfile , 'w')
         outf:write('')
         outf:close()
      end

      local outf = io.open(logfile, 'a')
      outf:write(string.format("%s %s  %s %s\n", os.date('%Y-%m-%d %H:%M:%S'), os.time(), direction, PDU))
      outf:close()
   end
end

-- pmMessage
local pmLuupTaskHandle = -1
function pmMessage(text, status)
   if (status == nil) then
      luup.task("", 4, nil, -1)
   else
      debug(text)
      pmLuupTaskHandle = luup.task(text, status, "PowerMax", pmLuupTaskHandle)
      if (status == 2) then
         luup.call_delay("pmMessage", 30, "", false)
      end
   end
end

-- Calculate CRC for PDU send to PowerMax 
function pmCalcCRC(pdu)
	local checksum = 0
	local newPDU
	local len = string.len(pdu)

	assert(len >= 2)
	for i = 2, (len - 2) do
	   checksum = checksum + tonumber(string.byte(pdu, i))
	end
	checksum = 0xFF - (checksum % 0xFF)
	if (checksum == 0xFF) then checksum = 0x00 end
	newPDU = string.sub(pdu, 1, len - 2) .. string.char(checksum, 0x0A)
	return newPDU
end

-- Check CRC in PDU received from PowerMax
function pmCheckCRC(pdu)
	local checksum = 0
	local len = string.len(pdu)

	assert(len >= 2)
	for i = 2, (len - 2) do
	   checksum = checksum + tonumber(string.byte(pdu, i))
	end
	checksum = 0xFF - (checksum % 0xFF)
	if (checksum == 0xFF) then checksum = 0x00 end
   local got = tonumber(string.byte(pdu, len - 1))
   if (checksum == got) then
      return true
   else
      debug(string.format("CRC check failed (expected %02X, got %02X)", checksum, got))
      return false
   end
end

-- pmString2Dword: convert 4 characters into an double-word integer
function pmString2Dword(s, start)
   return string.byte(s, start) + 0x100 * string.byte(s, start + 1) + 0x10000 * string.byte(s, start + 2) + 0x1000000 * string.byte(s, start + 3)
end

-- Convert PDU bytes to a string for logging 
function pmPduToString(pdu)
	 local PDUstr = ""
	 for i = 1, string.len(pdu) do
		PDUstr = PDUstr .. string.format("%02X ", string.byte(pdu, i))
	 end
	 return PDUstr
end

-- pmHandleCommException: we have run into a communication error
function pmHandleCommException(what)
   local exceptions = luup.variable_get(PANEL_SID, "CommExceptions", pmPanelDev)
   if (exceptions == nil) or (tonumber(exceptions) == nil) then
      exceptions = 0
   end
   exceptions = tonumber(exceptions) + 1
   luup.variable_set(PANEL_SID, "CommExceptions", exceptions, pmPanelDev)
   luup.variable_set(PANEL_SID, "CommLastException", os.time(), pmPanelDev)
   debug("*** Houston - we have a communication problem (" .. what .. ")! Executing a reload. ***")
   pmExpectedResponse = ""
   pmSendMessage("MSG_INIT")
   -- Execute a full Reload
   luup.inet.wget("http://localhost/port_3480/data_request?id=lu_reload")
end

-- pmPowerlinkEnrolled
function pmPowerlinkEnrolled()
   -- TODO: update DL messages (length) based on panel type
   pmMessage("Reading panel settings", 2)
   pmPowerlinkMode = true
   pmSendMessage("MSG_DL", { item = pmDownloadItem_t.MSG_DL_PANELFW })     -- Request the panel FW
   pmSendMessage("MSG_DL", { item = pmDownloadItem_t.MSG_DL_SERIAL })      -- Request serial & type (not always sent by default)
   pmSendMessage("MSG_DL", { item = pmDownloadItem_t.MSG_DL_ZONESTR })     -- Read the names of the zones
   pmSendMessage("MSG_START")      -- Start sending all relevant settings please
   local syncTime = (luup.variable_get(PANEL_SID, "AutoSyncTime", pmPanelDev) ~= "0")
   if (syncTime == true) then
      local t = os.date('*t')
      local year = t.year - 2000
      local timePdu = string.char(t.sec, t.min, t.hour, t.day, t.month, year)
      pmSendMessage("MSG_SETTIME", { time = timePdu })
   end
   pmSendMessage("MSG_EXIT")       -- Exit download mode
   luup.variable_set(PANEL_SID, "PowerlinkMode", "Powerlink", pmPanelDev)
end

-- pmIntervalCheck
function pmIntervalCheck(stuff)
   local now = os.time()
   -- Check Powerlink keep-alive
   if (pmPowerlinkMode == true) then
      local delta = now - (pmLastKeepAlive or 0)
      debug(string.format("Checking last keep alive (delta = %d)", delta))
      if (delta > 60) then
         -- Let Powermax know we are alive (and reset Powerlink communication error)
         debug("Clear Powerlink communication error")
         pmSendMessage("MSG_RESTORE")
         pmLastKeepAlive = now
         pmPowerlinkRetry = pmPowerlinkRetry + 1
         if (pmPowerlinkRetry == PL_RETRIES) then
            pmPowerlinkMode = false
            luup.variable_set(PANEL_SID, "PowerlinkMode", "Standard", pmPanelDev)
         end
      end
   end
   -- Check message sending timeout
   local timeout = (now - pmWaitingForResponse > TIMEOUT)
   if (timeout == true) and (pmExpectedResponse ~= "") then
      -- resend last message
      debug("*** Re-sending PDU (got no response) ***")
      if (pmSendMsgRetries == MSG_RETRIES) then
         pmHandleCommException("no response")
      else
         pmSendMsgRetries = pmSendMsgRetries + 1
         pmExpectedResponse = string.sub(pmExpectedResponse, 2)
         pmSendMessage(-1)
      end
   end
   -- Check if a tripped motion sensor has timed out
   for i = 1, 30 do
		local sensor = pmSensorDev_t[i]
      if (sensor ~= nil) then
         local child = sensor['id']
         local motion = (sensor['stype'] == "Motion")
         local lastTrip = luup.variable_get(SECURITY_SID, "LastTrip", child)
         if ((motion == true) and (lastTrip ~= nil)) and (now - lastTrip > (pmMotionOffDelay * 60)) then
            updateIfNeeded(SECURITY_SID, "Tripped", 0, child)
         end
      end
	end
   -- Check Siren
   if (pmSirenActive ~= nil) and (now > pmSirenActive) then
      pmSirenActive = nil
      for i = 1, #pmSirenDev_t do
         luup.variable_set(SIREN_SID, "Status", "0", pmSirenDev_t[i])
      end
   end
	luup.call_timer("pmIntervalCheck", 1, "20s", "", "")
end

-- pmCreateDevice
function pmCreateDevice(deviceList, altid, deviceName, zoneName, zoneType)
   local devVar = "VAR_" .. string.upper(deviceName)
   local devInit = pmDevices_t[devVar]

   assert(devInit ~= nil)
   local setVars = devInit[4]
   
   if (type(altid) == "number") then
      altid = string.format("%s-%d", deviceName, altid)
   elseif (string.sub(altid, 1, 1) == "Z") then
      local wirelessCnt = pmPanelConfig_t.CFG_WIRELESS[pmPanelTypeNr]
      local nr = tonumber(string.sub(altid, 2))
      if (nr > wirelessCnt) then
         setVars = CON_SENSOR_INIT -- wired zone
      end
   end
   if (zoneName == nil) or (zoneName == "") then
      zoneName = altid
   end
   if (zoneType ~= nil) and (zoneType ~= "") then
      zoneName = zoneName .. " (" .. zoneType .. ")"
   else
      zoneName = zoneName .. " (PowerMax)"
   end
   if (fileExists(devInit[2]) == false) then
      debug("File " .. devInit[2] .. " does not exist. Cannot create " .. altid)
      return nil
   else
      pmMessage("Adding " .. deviceName .. " device " .. altid .. " (" .. zoneName .. ")", 4)
      luup.chdev.append(pmPanelDev, deviceList, altid, zoneName, devInit[1], devInit[2], devInit[3], setVars, devInit[5])
      return altid
   end
end

-- Initialises the panel, handlers etc.
function pmStartup(lul_device)
	pmPanelDev = lul_device

   local exceptions = luup.variable_get(PANEL_SID, "CommExceptions", pmPanelDev)
   if (exceptions ~= nil) then
      exceptions = tonumber(exceptions, 10)
   end
   if (exceptions ~= nil) and (exceptions > MAX_EXCEPTIONS) then
      local last = tonumber(luup.variable_get(PANEL_SID, "CommLastException", pmPanelDev))
      local diff = os.difftime(os.time(), last)
      if (diff < 86400) then -- more than a day ago?
         pmMessage("Not starting due to too many communcation errors.", 2)
         return false
      else
         -- reset communication error count
         luup.variable_set(PANEL_SID, "CommExceptions", "0", pmPanelDev)
      end
   end

	luup.log("POWERMAX: starting ... device #" .. tostring(pmPanelDev), 10)

   for i = 1, #pmPanelInit_t do
      	updateIfNeeded(PANEL_SID, pmPanelInit_t[i][1], pmPanelInit_t[i][2], pmPanelDev, pmPanelInit_t[i][3])
   end

   pmLogDebug = (luup.variable_get(PANEL_SID, "PluginDebug", pmPanelDev) ~= "0")
   local forceStandard = (luup.variable_get(PANEL_SID, "ForceStandard", pmPanelDev) == "1")
   pmLang = luup.variable_get(PANEL_SID, "PluginLanguage", pmPanelDev)
   pmRemoteArm = luup.variable_get(PANEL_SID, "EnableRemoteArm", pmPanelDev)
   pmSensorArm = luup.variable_get(PANEL_SID, "SensorArm", pmPanelDev)
   pmMotionOffDelay = tonumber(luup.variable_get(PANEL_SID, "MotionOffDelay", pmPanelDev), 10)
   
	if (luup.io.is_connected(pmPanelDev) == false) then
		pmMessage("Please select the Serial Port for the PowerMax", 2)
		return false
	end

   -- Register all message handlers
   pmReceiveMsg_t[0x02][2] = pmHandleAck
   pmReceiveMsg_t[0x06][2] = pmHandleTimeout
   pmReceiveMsg_t[0x08][2] = pmHandleDenied
   pmReceiveMsg_t[0x0B][2] = pmHandleStop
   pmReceiveMsg_t[0x33][2] = pmHandleSettings
   pmReceiveMsg_t[0x3C][2] = pmHandleInfo
   pmReceiveMsg_t[0x3F][2] = pmHandleSettingsItem
   pmReceiveMsg_t[0xA0][2] = pmHandleEventlog
   pmReceiveMsg_t[0xA5][2] = pmHandleStatus
   pmReceiveMsg_t[0xA7][2] = pmHandlePanel
   if (forceStandard == true) then
      pmReceiveMsg_t[0xAB] = nil
   else
      pmReceiveMsg_t[0xAB][2] = pmHandlePowerlink
   end

   pmSendMessage("MSG_INIT")
   if (forceStandard == false) then
      pmStartDownload()
   else
      pmProcessSettings()
      luup.variable_set(PANEL_SID, "PowerlinkMode", "Standard", pmPanelDev)
   end
   
	luup.call_timer("pmIntervalCheck", 1, "20s", "", "")
end

-- pmGetPin: Convert a PIN given as string in the PIN PDU format as used in messages to powermax
function pmGetPin(pin) 
	local pinPDU

	if (pin == "") or (pin == nil) or (string.len(pin) ~= 4) then
      if (pmPowerlinkMode == true) then
         return pmPincode_t[1], true
      else
         pmMessage("4 digit PIN needed", 2)
         pinPDU = string.char(0x00, 0x00)
         return pinPDU, false
      end
	end
	pinPDU = string.char(tonumber(string.sub(pin, 1, 2), 16), tonumber(string.sub(pin, 3, 4), 16))

	return pinPDU, false
end

-- pmWriteSettings: add a certain setting to the settings table
function pmWriteSettings(page, index, setting)
   local len = string.len(setting)
   local wrap = (index + len - 0x100)
   local sett = { "", "" }
   local i
   assert(len <= 0xB1)
   if (wrap > 0) then
      sett[1] = string.sub(setting, 1, len - wrap)
      sett[2] = string.sub(setting, len - wrap + 1, len + 1)
      wrap = 1
   else
      sett[1] = setting
      wrap = 0
   end
   for i = 0, wrap do
      if (pmSettings_t[page + i] == nil) then
         pmSettings_t[page + i] = string.rep(string.char(0xFF), 0x100)
      end
      len = string.len(sett[i + 1]) + 1
      if (i == 1) then 
         index = 0 
      end
      pmSettings_t[page + i] = string.sub(pmSettings_t[page + i], 1, index) .. sett[i + 1] .. string.sub(pmSettings_t[page + i], index + len, 0x100)
   end
end

-- pmReadSettings
function pmReadSettings(item)
   local index = string.byte(item, 1) + 1
   local page = string.byte(item, 2)
   local len = string.byte(item, 3) + 0x100 * string.byte(item, 4)
   local s = ""
   if (pmSettings_t[page] ~= nil) then
      while (len > 0) do
         local str = string.sub(pmSettings_t[page], index, index + len - 1)
         s = s .. str
         page = page + 1
         index = 1
         len = len - string.len(str)
      end
   end
   return s
end

-- pmProcessSetting
function pmProcessSettings()
   local i, setting
   local x10_t = {}
   local doorZoneStr = ""
   local motionZoneStr = ""
   local smokeZoneStr = ""
   local deviceStr = ""
   local childDevices = luup.chdev.start(pmPanelDev)
   local autoCreate = (luup.variable_get(PANEL_SID, "AutoCreate", pmPanelDev) == "1")
   local devices = luup.variable_get(PANEL_SID, "Devices", pmPanelDev)

   local panelSerialType = pmReadSettings(pmDownloadItem_t.MSG_DL_SERIAL)
   if (panelSerialType ~= "") then
      pmPanelTypeNr = string.byte(panelSerialType, 8) + 1
      local model = pmPanelType_t[pmPanelTypeNr] or "UNKNOWN"
      updateIfNeeded(PANEL_SID, "PanelType", model, pmPanelDev)
      pmDumpSettings()
   end
   
   local zoneCnt = pmPanelConfig_t.CFG_WIRELESS[pmPanelTypeNr] + pmPanelConfig_t.CFG_WIRED[pmPanelTypeNr]
   local customCnt = pmPanelConfig_t.CFG_ZONECUSTOM[pmPanelTypeNr]
   local userCnt = pmPanelConfig_t.CFG_USERCODES[pmPanelTypeNr]
   local partitionCnt = pmPanelConfig_t.CFG_PARTITIONS[pmPanelTypeNr]
   local sirenCnt = pmPanelConfig_t.CFG_SIRENS[pmPanelTypeNr]
   local keypad1wCnt = pmPanelConfig_t.CFG_1WKEYPADS[pmPanelTypeNr]
   local keypad2wCnt = pmPanelConfig_t.CFG_2WKEYPADS[pmPanelTypeNr]

   for i = 1, 1 do -- TODO: '1 + partitionCnt' for multiple partition support
      pmCreateDevice(childDevices, i, "Partition")
      local s = string.format("Partition-%d", i)
      pmPartitionDev_t[i] = findChild(pmPanelDev, s)
   end
   
   if (pmPowerlinkMode == true) then
      debug("Processing settings information")
      -- Process zone names of this panel
      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_ZONESTR)
      for i = 1, 26 + customCnt do
         local s = string.sub(setting, (i - 1) * 0x10 + 1, i * 0x10):gsub("%s*$", "")
         if (string.byte(s, 1) ~= 0xFF) then
            pmZoneName_t[i] = string.sub(s, 1, 1) .. string.lower(string.sub(s, 2, 0x0F))
         end
      end

      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_COMMDEF)
      pmBellTime = string.byte(setting, 0x04)
      pmSilentPanic = (bitw.band(string.byte(setting, 0x1A), 0x10) == 0x10)
      pmQuickArm = (bitw.band(string.byte(setting, 0x1B), 0x08) == 0x08)
      pmBypassOff = (bitw.band(string.byte(setting, 0x1C), 0xC0) == 0)
      pmForcedDisarmCode = string.sub(setting, 0x11, 0x12)
      debug(string.format("Bell time = %s minutes; Bypass off = %s; Quick arm = % s; Silent panic = %s", pmBellTime, tostring(pmBypassOff), tostring(pmQuickArm), tostring(pmSilentPanic)))
      if (pmSensorArm == "auto") then
         pmSensorShowBypass = not(pmBypassOff)
      elseif (pmSensorArm == "bypass") then
         pmSensorShowBypass = true
      elseif (pmSensorArm == "live") then
         pmSensorShowBypass = false
      end
         
      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_PINCODES)
      for i = 1, userCnt do
         pmPincode_t[i] = string.sub(setting, 2 * i - 1, 2 * i)
      end

      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_PANELFW)
      local panelEprom = string.sub(setting, 0x01, 0x10)
      local panelSoftware = string.sub(setting, 0x11, 0x20)
      debug(string.format("EPROM: %s; SW: %s", panelEprom, panelSoftware))
      updateIfNeeded(PANEL_SID, "PanelEprom", panelEprom, pmPanelDev)
      updateIfNeeded(PANEL_SID, "PanelSoftware", panelSoftware, pmPanelDev)

      local idx = string.format("%02x%02x", string.byte(panelSerialType, 8), string.byte(panelSerialType, 7))
      local pmPanelName = pmPanelName_t[idx] or "UNKNOWN"
      local pmPanelSerial = ""
      for i = 0, 5 do
         local nr = string.byte(panelSerialType, i + 1)
         local s = (nr == 0xFF) and "." or string.format("%02X", nr)
         pmPanelSerial = pmPanelSerial .. s
      end
      debug(string.format("Panel %s with serial %s", pmPanelName, pmPanelSerial))
      updateIfNeeded(PANEL_SID, "PanelName", pmPanelName, pmPanelDev)
      updateIfNeeded(PANEL_SID, "PanelSerial", pmPanelSerial, pmPanelDev)

      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_ZONES)
      local zoneNames = pmReadSettings(pmDownloadItem_t.MSG_DL_ZONENAMES)
      for i = 1, zoneCnt do
         local zoneEnrolled = (string.sub(setting, i * 4 - 3, i * 4 - 1) ~= string.char(0, 0, 0))
         local zoneName = pmZoneName_t[string.byte(zoneNames, i) + 1]
         if (zoneEnrolled == true) then
            local zoneInfo = string.byte(setting, i * 4)
            local zoneType = bitw.band(zoneInfo, 0x0F) + 1
            local zoneChime = bitw.rshift(zoneInfo, 4) + 1
            local sensorId = string.byte(setting, i * 4 - 1)
            local sensorType = pmZoneSensor_t[bitw.band(sensorId, 0xF)] or "UNKNOWN"
            debug(string.format("Zone %02d: %s (%s; sensor type = %02X [%s])" , i, pmZoneType_t[pmLang][zoneType], pmZoneChime_t[zoneChime], sensorId, sensorType))
            pmSensorDev_t[i] = { stype = sensorType, sid = sensorId, ztype = zoneType, zname = zoneName, zchime = zoneChime }
            if (sensorType == "Magnet") then
               doorZoneStr = string.format("%s,Z%02d", doorZoneStr, i)
            elseif (sensorType == "Motion") then
               motionZoneStr = string.format("%s,Z%02d", motionZoneStr, i)
            elseif (sensorType == "Smoke") or (sensorType == "Gas") then
               smokeZoneStr = string.format("%s,Z%02d", smokeZoneStr, i)
            end
         end
      end

      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_PGMX10)
      local x10Names = pmReadSettings(pmDownloadItem_t.MSG_DL_X10NAMES)
      for i = 0, 15 do
         local enabled = false
         local x10Name = 0x1F
         for j = 0, 8 do
            enabled = enabled or (string.byte(setting, 6 + i + j * 0x10) ~= 0)
         end
         if (i > 0) then
            x10Name = string.byte(x10Names, i)
            x10_t[i] = pmZoneName_t[x10Name + 1]
         end
         if (enabled == true) or (x10Name ~= 0x1F) then
            if (i == 0) then
               deviceStr = deviceStr .. ",PGM"
            else
               deviceStr = string.format("%s,X%02d", deviceStr, i)
               if (string.find(devices, string.format("X%02dd", i)) ~= nil) then
                  deviceStr = deviceStr .. "d"
               end
            end
         end
      end

      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_1WKEYPAD)
      for i = 1, keypad1wCnt do
         local keypadEnrolled = (string.sub(setting, i * 4 - 3, i * 4 - 2) ~= string.char(0, 0))
         if (keypadEnrolled == true) then
            deviceStr = string.format("%s,K1%d", deviceStr, i)
         end
      end
      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_2WKEYPAD)
      for i = 1, keypad2wCnt do
         local keypadEnrolled = (string.sub(setting, i * 4 - 3, i * 4 - 1) ~= string.char(0, 0, 0))
         if (keypadEnrolled == true) then
            deviceStr = string.format("%s,K2%d", deviceStr, i)
         end
      end

      setting = pmReadSettings(pmDownloadItem_t.MSG_DL_SIRENS)
      for i = 1, sirenCnt do
         local sirenEnrolled = (string.sub(setting, i * 4 - 3, i * 4 - 1) ~= string.char(0, 0, 0))
         if (sirenEnrolled == true) then
            deviceStr = string.format("%s,S%02d", deviceStr, i)
         end
      end
   end

	pmMessage("Adding zone devices", 4)
   -- Use variables if AutoCreate is false or not in Powerlink mode, otherwise use what we read in the settings
   local doorZones = doorZoneStr
   local motionZones = motionZoneStr
   local smokeZones = smokeZoneStr
   devices = deviceStr
   if (autoCreate == false) or (pmPowerlinkMode == false) then
      doorZones = luup.variable_get(PANEL_SID, "DoorZones", pmPanelDev)
      motionZones = luup.variable_get(PANEL_SID, "MotionZones", pmPanelDev)
      smokeZones = luup.variable_get(PANEL_SID, "SmokeZones", pmPanelDev)
      devices = luup.variable_get(PANEL_SID, "Devices", pmPanelDev)
   elseif (pmPowerlinkMode == true) then
      updateIfNeeded(PANEL_SID, "DoorZones", string.sub(doorZoneStr, 2), pmPanelDev)
      updateIfNeeded(PANEL_SID, "MotionZones", string.sub(motionZoneStr, 2), pmPanelDev)
      updateIfNeeded(PANEL_SID, "SmokeZones", string.sub(smokeZoneStr, 2), pmPanelDev)
      updateIfNeeded(PANEL_SID, "Devices", string.sub(deviceStr, 2), pmPanelDev)
   end
   
   -- Start adding zones
   local zoneName, zoneTypeName
   for i = 1, zoneCnt do
      local s = string.format("Z%02d", i)
      if (pmSensorDev_t[i] ~= nil) then
         zoneName = pmSensorDev_t[i]['zname']
         zoneTypeName = pmZoneType_t[pmLang][pmSensorDev_t[i]['ztype']]
      else
         zoneName = nil
         zoneTypeName = nil
      end
      local stype = nil
      if (string.find(doorZones, s) ~= nil) then
         stype = "Magnet"
      elseif (string.find(motionZones, s) ~= nil) then
         stype = "Motion"
      elseif (string.find(smokeZones, s) ~= nil) then
         stype = "Smoke"
      end
      local sensor = nil
      if (stype ~= nil) then
         sensor = pmCreateDevice(childDevices, s, stype, zoneName, zoneTypeName)
      end
      if (sensor ~= nil) then
         if (pmSensorDev_t[i] == nil) then
            pmSensorDev_t[i] = {}
         end
         pmSensorDev_t[i].id = findChild(pmPanelDev, sensor)
         pmSensorDev_t[i].name = sensor
         pmSensorDev_t[i].stype = stype
         pmSensorDev_t[i].partition = { 1 }
      end
   end
	
   -- Add PGM and X10 devices
	for i = 0, 15 do
		if (i == 0) then
			if (string.find(devices, "PGM")) then
            pmCreateDevice(childDevices, "PGM", "Switch", "PGM")
			end
		else
			s = string.format("X%02d", i)
         local pos = string.find(devices, s)
			if (pos ~= nil) then
            local devInit
            if (string.sub(devices, pos + 3, pos + 3) == "d") then
               pmCreateDevice(childDevices, s, "Dim", x10_t[i], s)
            else
               pmCreateDevice(childDevices, s, "Switch", x10_t[i], s)
            end
			end
		end
		local x10_device = findChild(pmPanelDev, s)
		if (x10_device ~= nil) then
         x10_dev_nr = string.format("%04x", 2 ^ i)
         updateIfNeeded(PANEL_SID, "X10DeviceNr", x10_dev_nr, x10_device, true)
		end
	end
   
   -- Add sirens
   for i = 1, sirenCnt do
      local sirenEnrolled = (string.find(devices, string.format("S%02d", i)) ~= nil)
      if (sirenEnrolled == true) then
         local id = pmCreateDevice(childDevices, i, "Siren")
         if (id ~= nil) then
            pmSirenDev_t[i] = findChild(pmPanelDev, id)
         end
      end
   end

   -- Add keypads
   for i = 1, keypad1wCnt do
      local keypadEnrolled = (string.find(devices, string.format("K1%d", i)) ~= nil)
      if (keypadEnrolled == true) then
         pmCreateDevice(childDevices, i, "Keypad", nil, "1-way")
      end
   end
   
   for i = 1, keypad2wCnt do
      local keypadEnrolled = (string.find(devices, string.format("K2%d", i)) ~= nil)
      if (keypadEnrolled == true) then
         pmCreateDevice(childDevices, i, "Keypad", nil, "2-way")
      end
   end
   luup.chdev.sync(pmPanelDev, childDevices)
   pmSendMessage("MSG_STATUS")
   pmMessage("Ready for use", 2)
end

-- pmSendPdu: Send the pdu to the PowerMax via serial
function pmSendPdu(pdu)
   sendPdu = pmCalcCRC(pdu)
   local PDUstr = pmPduToString(sendPdu)
   debug(string.format("PDU sent to panel: %s", PDUstr))
   pmLogPdu(PDUstr, "-VR->")
   if luup.io.write(sendPdu) == false then
      luup.log("POWERMAX: cannot send PDU", 2)
      luup.set_failure(true)
      return ""
   end
   return sendPdu
end

-- pmSendAck: Send full or partial acknowledge to Powermax (depending on operating mode)
function pmSendAck(ackType)
   local lastType = string.byte(pmLastPDU, 2)
   local normalMode = (lastType >= 0x80) or ((lastType < 0x10) and (string.byte(pmLastPDU, string.len(pmLastPDU) - 2) == 0x43))
   
   if (normalMode == true) then
      pmSendPdu(string.char(0x0D, ackType, 0x43, 0x00, 0x0A), 0)
   else
      pmSendPdu(string.char(0x0D, ackType, 0x00, 0x0A), 0)
   end
end

-- pmStartDownload (exposed as called with delay)
function pmStartDownload(stuff)
   pmSendMessage("MSG_DOWNLOAD")	-- If we get a NACK, then the download code is not valid (not enrolled)
   pmLastKeepAlive = os.time()
end

-- pmSendMessage: Send (known) messages to the PowerMax via serial
-- Returns success: true: sent or queued; false: nothing sent 
function pmSendMessage(msg, modifiers)
   local outPdu
   local response
   local interval = os.time() - pmWaitingForResponse
   local timeout = (interval > TIMEOUT)

   if (msg == nil) then -- get message from queue
      local tab = Queue.popleft(pmOutgoingQueue)
      if (tab == nil) then
         -- nothing to send
         return false
      end
      debug("*** Sending next PDU in queue ***")
      outPdu = tab[1]
      response = tab[2]
   elseif (msg == -1) then -- re-send last message
      outPdu = pmLastSendMsg
      response = pmExpectedResponse
      timeout = true
   else
      msg = pmSendMsg_t[msg]
      assert(msg ~= nil)
      outPdu = string.char(0x0D) .. msg[1] .. string.char(0x00, 0x0A)
      response = (msg[2] == nil) and "" or string.char(msg[2])
   
      if (modifiers ~= nil) then
         for fnd, rpl in pairs(modifiers) do
            rpl = string.gsub(rpl, "%%", "%%%%")
            outPdu = string.gsub(outPdu, fnd, rpl, 1)
         end
      end
      
      if (string.byte(outPdu, 2) == 0x3E) then
         local cnt = (string.byte(outPdu, 5)) + 0x100 * string.byte(outPdu, 6)
         response = string.rep(response, math.floor(cnt / 0xB0) + 1)
      end
   end
   
	if (timeout == true) or (pmExpectedResponse == "") then -- we are ready to send
      -- Always expect an ACK (ACK send messages don't pass through here)
      pmExpectedResponse = string.char(0x02) .. response
      if (interval < 1) then
         debug("Wait before sending next command")
         luup.sleep(500)
      end
      pmLastPDU = pmSendPdu(outPdu)
      pmLastSendMsg = pmLastPDU
      if (pmLastPDU == "") then
         return false
      end
      pmWaitingForResponse = os.time()
	else -- queue message
		Queue.pushright(pmOutgoingQueue, { outPdu, response })
	end
	return true
end

-- A byte of data has been received via serial
-- This function is invoked until a complete PDU is assembled
function pmIncoming(data)
   local pduLen = string.len(pmIncomingPdu)
   local success
   
	if (pduLen == 0) then
		if (string.byte(data) == 0x0D) then -- preamble
			debug("start of new PDU detected")
			pmIncomingPdu = data
      end
   elseif (pduLen == 1) then
      local msgType_t = pmReceiveMsg_t[string.byte(data)]
      pmIncomingPduLen = (msgType_t ~= nil) and msgType_t[1] or 0
      debug(string.format("Message %02X; pmIncomingPduLen = %d", string.byte(data), pmIncomingPduLen))
      pmIncomingPdu = pmIncomingPdu .. data
	elseif (pmIncomingPduLen == 0 and string.byte(data) == 0x0A) or (pduLen + 1 == pmIncomingPduLen) then -- postamble
		pmIncomingPdu = pmIncomingPdu .. data
		if (pmCheckCRC(pmIncomingPdu) == true) then
         pmLastPDU = pmIncomingPdu
			local PDUstr = pmPduToString(pmIncomingPdu)
         local msgType = string.byte(pmIncomingPdu, 2)
         local msgType_t = pmReceiveMsg_t[msgType]
			debug(string.format("PDU received %s", PDUstr))
			pmLogPdu(PDUstr, "<-PM-")
         pmWaitingForResponse = os.time()
         
         if (msgType_t == nil) then
            debug(string.format("Unhandled message %02X", msgType))
            pmSendAck(0x02)
         else
            -- Send an ACK if needed
            if (msgType_t[3] == true) then
               pmSendAck(0x02)
            end
            -- Handle the message
            local msgHandler = msgType_t[2] or 0
            if (msgHandler == 0) then
               debug(string.format("Unhandled message %02X", msgType))
            else
               success = msgHandler()
            end
            -- Check response
            if (pmExpectedResponse ~= "") then
               -- We've sent something and are waiting for a reponse - this is it
               local firstExpectedResponse = string.byte(pmExpectedResponse, 1)
               pmExpectedResponse = string.sub(pmExpectedResponse, 2)
               if (firstExpectedResponse ~= msgType) then
                  if (pmSendMsgRetries == 0) then
                     debug(string.format("*** Waiting for next PDU (expected %02X got %02X) ***", firstExpectedResponse, msgType))
                     pmSendMsgRetries = 1 -- don't resend immediately - first wait for the next response
                     pmExpectedResponse = string.char(firstExpectedResponse) .. pmExpectedResponse
                  else
                     debug(string.format("*** Re-sending PDU (expected %02X got %02X) ***", firstExpectedResponse, msgType))
                     if (pmSendMsgRetries == MSG_RETRIES) then
                        pmHandleCommException("wrong response")
                     else
                        pmSendMsgRetries = pmSendMsgRetries + 1
                        pmSendMessage(-1)
                     end
                  end
               else
                  pmSendMsgRetries = 0
               end
            end
         end
         if (pmExpectedResponse == "") then
            pmSendMessage(nil)
			end
			pmIncomingPdu = ""
      else -- CRC check failed
         if (pmIncomingPduLen > 0) then
            pmCrcErrorCount = pmCrcErrorCount + 1
            if (pmCrcErrorCount > MAX_CRC_ERROR) then
               pmHandleCommException("CRC errors")
            else
               local PDUstr = pmPduToString(pmIncomingPdu)
               debug(string.format("PDU received %s", PDUstr))
               pmLogPdu(PDUstr, "<-PM-")
               pmWaitingForResponse = os.time()
               pmIncomingPdu = ""
               pmSendMessage("MSG_INIT") -- Re-initialize
            end
         end
      end
	else
		pmIncomingPdu = pmIncomingPdu .. data
	end
	return true
end

-- pmHandleAck (0x02)
function pmHandleAck()
   local lastSendMsgType = string.byte(pmLastSendMsg, 2)

   if (lastSendMsgType == 0x0F) then
      pmProcessSettings()
   end
   return true
end

-- pmHandleTimeout (0x06)
function pmHandleTimeout()
   debug("Powermax sends timeout")
   pmExpectedResponse = string.char(0x06)
   pmSendMessage("MSG_EXIT")
   return true
end

-- pmHandleDenied (0x08)
function pmHandleDenied()
   local lastSendMsgType = string.byte(pmLastSendMsg, 2)
   debug(string.format("Handling DENIED message (%02X)", lastSendMsgType))
   
   if (bitw.band(lastSendMsgType, 0xA0) == 0xA0) then -- A0, A1, AA
      pmMessage("Wrong PIN entered.", 2)
      if (lastSendMsgType == 0xA0) then
         luup.variable_set(PANEL_SID, "EventLog", "*** Access denied ***", device)
      end
   elseif (lastSendMsgType == 0x24) then
      pmPowerlinkMode = false
      luup.variable_set(PANEL_SID, "PowerlinkMode", "Standard", pmPanelDev)
      pmMessage("Vera Powerlink not enrolled.", 2)
   end
   pmExpectedResponse = string.char(0x08) .. string.sub(pmExpectedResponse, 2)
   return true
end

-- pmHandleStop (0x0B)
function pmHandleStop()
   -- We're not expecting any more responses
   pmExpectedResponse = ""
   return true
end

-- pmHandleSettings (0x33)
function pmHandleSettings()
   local index = string.byte(pmIncomingPdu, 3)
   local page = string.byte(pmIncomingPdu, 4)
   local lastSendMsgType = string.byte(pmLastSendMsg, 2)
   debug(string.format("Received Powermax setting page %02X index %02X", page, index))
   if (lastSendMsgType == 0x0A) then
      pmExpectedResponse = pmExpectedResponse .. string.char(0x33) -- we're not done; keep them coming
   end
   pmWriteSettings(page, index, string.sub(pmIncomingPdu, 5, 12))
   return true
end

-- pmHandleInfo (0x3C)
function pmHandleInfo()
   pmPanelTypeNr = string.byte(pmIncomingPdu, 8) + 1
   local model = pmPanelType_t[pmPanelTypeNr] or "UNKNOWN"
   updateIfNeeded(PANEL_SID, "PanelType", model, pmPanelDev)
   pmPowerlinkEnrolled()
   return true
end

-- pmHandleSettingsItem (0x3F)
function pmHandleSettingsItem()
   local idx = string.byte(pmIncomingPdu, 3)
   local page = string.byte(pmIncomingPdu, 4)
   local len = string.byte(pmIncomingPdu, 5)
   debug(string.format("Received Powermax setting page %02X index %02X", page, idx))
   pmWriteSettings(page, idx, string.sub(pmIncomingPdu, 6, 5 + len))
end

-- pmHandleEventlog (0xA0)
function pmHandleEventlog()
   debug(string.format("eventlog received"))
   local eventNum = string.byte(pmIncomingPdu, 4)
   if (eventNum == 0x01) then
      pmEventLog = ""
      pmEventCnt = string.byte(pmIncomingPdu, 3)
      pmExpectedResponse = pmExpectedResponse .. string.char(0xA0)
   else
      local secNum = string.byte(pmIncomingPdu, 5)
      local minNum = string.byte(pmIncomingPdu, 6)
      local horNum = string.byte(pmIncomingPdu, 7)
      local dayNum = string.byte(pmIncomingPdu, 8)
      local monNum = string.byte(pmIncomingPdu, 9)
      local yerNum = 2000 + string.byte(pmIncomingPdu, 10)
      local eventZone = string.byte(pmIncomingPdu, 11)
      local logEvent = string.byte(pmIncomingPdu, 12)
      local eventType = bitw.band(logEvent, 0x7F)
      local isPanelEvent = (bitw.band(logEvent, 0x80) == 0x80)
      pmEventLog = pmEventLog .. string.format("\r\n%02d/%02d/%d %02d:%02d:%02d    ", dayNum, monNum, yerNum, horNum, minNum, secNum)
      pmEventLog = pmEventLog .. (pmLogUser_t[eventZone + 1] or "UNKNOWN")
      if (isPanelEvent == true) then
         pmEventLog = pmEventLog .. "  * "
      else
         pmEventLog = pmEventLog .. "    "
      end
      pmEventLog = pmEventLog .. (pmLogEvent_t[pmLang][eventType + 1] or "UNKNOWN")
      if (eventNum == pmEventCnt) then
         luup.variable_set(PANEL_SID, "EventLog", pmEventLog, pmPanelDev)
      else
         pmExpectedResponse = pmExpectedResponse .. string.char(0xA0)
      end
   end
   return true
end

-- pmHandleStatus (0xA5)
function pmHandleStatus()
   local msgTot = string.byte(pmIncomingPdu, 3)
   local eventType = string.byte(pmIncomingPdu, 4)
   if (msgTot > 0) and (eventType ~= msgTot) then
      pmExpectedResponse = pmExpectedResponse .. string.char(0xA5) -- we're not done; keep them coming
   end
   if (eventType == 0x02) then
      debug("Received zone status message")
      local zoneStatus = pmString2Dword(pmIncomingPdu, 5)
      local batteryStatus = pmString2Dword(pmIncomingPdu, 9)
      for i = 1, 30 do
         local sensor = pmSensorDev_t[i]
         local trip = bitw.band(zoneStatus, 2 ^ (i - 1)) > 0 and 1 or 0
         if (sensor ~= nil) then
            if (updateIfNeeded(SECURITY_SID, "Tripped", trip, sensor['id'])) then
               debug(string.format("Updating tripped status for zone device %02d", i))
            end
            updateIfNeeded(SENSOR_SID, "BatteryLevel", bitw.band(batteryStatus, 2 ^ (i - 1)) > 0 and 5 or 100, sensor['id'])
         else
            if (trip == 1) then
               debug("Sensor zone " .. i .. " tripped, but no device found!")
            end
         end
      end
   elseif (eventType == 0x04) then
      local sysStatus = string.byte(pmIncomingPdu, 5)
      local sysFlags  = string.byte(pmIncomingPdu, 6)
      local eventZone = string.byte(pmIncomingPdu, 7)
      local zoneEType = string.byte(pmIncomingPdu, 8)
      local x10Stat1  = string.byte(pmIncomingPdu, 11)
      local x10Stat2  = string.byte(pmIncomingPdu, 12)

      if (zoneEType ~= 0x00) then
         debug(string.format("Event %s in Zone %02d", (pmEventType_t[pmLang][zoneEType+1] or "UNKNOWN"), eventZone))
         local sensor = pmSensorDev_t[eventZone]
         if (sensor == nil) then
            debug(string.format("unable to locate zone device %02d", eventZone))
         else
            local child = sensor['id']
            if (zoneEType == 0x03) then
               updateIfNeeded(SECURITY_SID, "Tripped", 1, child) -- prevent updating twice; status msg comes before event msg
               luup.variable_set(SECURITY_SID, "LastTrip", os.time(), child)
            elseif (zoneEType == 0x04) then
               updateIfNeeded(SECURITY_SID, "Tripped", 0, child)
            elseif (zoneEType == 0x05) then
               if (updateIfNeeded(SECURITY_SID, "Tripped", 1, child) == true) then
                  luup.variable_set(SECURITY_SID, "LastTrip", os.time(), child)
               end
            end
         end                    
      end

      -- update status of PGM
      local child = findChild(pmPanelDev, string.format("PGM"))
      if (child ~= nil) then
         if (bitw.band(x10Stat1, 1) == 1) then
            if (updateIfNeeded(SECURITY_SID, "Tripped", 1, child) == true) then
               luup.variable_set(SECURITY_SID, "LastTrip", os.time(), child)
            end
         else
            updateIfNeeded(SECURITY_SID, "Tripped", 0, child)
         end
      end

      -- update status of X10 devices
      local x10Status = x10Stat1 + (x10Stat2 ^ 8)
      for i = 1, 15 do
         local s = string.format("X%02d", i)
         child = findChild(pmPanelDev, s)
         if (child ~= nil) then
            local status = bitw.band(x10Status, 2 ^ i) > 0 and 1 or 0
            if (updateIfNeeded(X10DEVICE_SID, "Status", status, child) == true) then
               debug("Updating status of X10 device " .. s .. " (status = " .. status .. ")")
            end
         end
      end

      local sysFlagsStr = " "
      if (bitw.band(sysFlags, 1) == 0) then 
         sysFlagsStr = sysFlagsStr .. ",NRDY" 
      end
      if (bitw.band(sysFlags, 2) == 2) then 
         sysFlagsStr = sysFlagsStr .. ",MEM"
         updateIfNeeded(PARTITION_SID, "AlarmMemory", 1, pmPartitionDev_t[1])
      else
         updateIfNeeded(PARTITION_SID, "AlarmMemory", 0, pmPartitionDev_t[1])
      end
      if (bitw.band(sysFlags, 4) == 4) then 
         sysFlagsStr = sysFlagsStr .. ",TRBL"
         pmMessage("Trouble - check system.", 2)
      end
      if (bitw.band(sysFlags, 128) == 128) then 
         sysFlagsStr = sysFlagsStr .. ",ALRM" 
         if(updateIfNeeded(PARTITION_SID, "Alarm", "Active", pmPartitionDev_t[1]) == true) then
            luup.variable_set(PARTITION_SID, "LastAlarmActive", os.time(), pmPartitionDev_t[1])
         end
      else
         updateIfNeeded(PARTITION_SID, "Alarm", "None", pmPartitionDev_t[1])
      end
      local armModeNum = (pmArmed_t[sysStatus] ~= nil) and 1 or 0
      local armMode = (armModeNum == 1) and "Armed" or "Disarmed"
      local s = ""
      for i = 0, 7 do
         if (bitw.band(sysFlags, 2 ^ i) ~= 0) then
            if (i == 5) then
               if (eventZone == 0xFF) then
                  s = s .. (pmEventType_t[pmLang][zoneEType + 1] or "UNKNOWN") .. " from Panel, "
               else
                  s = s .. string.format("%s in Zone %02d, ", (pmEventType_t[pmLang][zoneEType + 1] or "UNKNOWN"), eventZone)
               end
            else
               s = s .. (pmSysStatusFlags_t[pmLang][i + 1] or "UNKNOWN") .. ", "
            end
         end
      end
      s = (s == "") and "Not ready" or string.sub(s, 1, string.len(s) - 2)
      debug(string.format("System state is %s %s",(pmSysStatus_t[pmLang][sysStatus + 1] or "UNKNOWN"), sysFlagsStr))
      updateIfNeeded(PARTITION_SID, "VendorStatus", (pmSysStatus_t[pmLang][sysStatus + 1]  or "UNKNOWN") .. sysFlagsStr, pmPartitionDev_t[1])
      updateIfNeeded(PARTITION_SID, "VendorStatusCode", string.format("%02X%02X", sysStatus, sysFlags), pmPartitionDev_t[1])
      updateIfNeeded(PARTITION_SID, "VendorStatusData", s, pmPartitionDev_t[1])
		updateIfNeeded(PARTITION_SID, "ArmMode", armMode, pmPartitionDev_t[1])
		updateIfNeeded(PARTITION_SID, "ArmModeNum", armModeNum, pmPartitionDev_t[1])
      updateIfNeeded(PARTITION_SID, "DetailedArmMode", (pmDetailedArmMode_t[sysStatus + 1] or "UNKNOWN"), pmPartitionDev_t[1])
      if (pmSensorShowBypass == false) then
         for i = 1, 30 do
            local sensor = pmSensorDev_t[i]
            if (sensor ~= nil) then
               local zoneType = sensor['ztype'] - 1
               local mode = bitw.band(sysStatus, 0x0F)
               armed = ((zoneType > 0) and ((bitw.band(zoneType, 0x14) == 0) or (mode == 0x5) or ((mode == 0x4) and (zoneType % 6 ~= 0)))) and "1" or "0"
               updateIfNeeded(SECURITY_SID, "Armed", armed, sensor['id'])
            end
         end
      end
   elseif (eventType == 0x06) then
      debug("Received zone bypass message")
      --local zoneEnrolled = pmString2Dword(pmIncomingPdu, 5)
      if (pmSensorShowBypass == true) then
         local zoneBypass = pmString2Dword(pmIncomingPdu, 9)
         for i = 1, 30 do
            local sensor = pmSensorDev_t[i]
            if (sensor ~= nil) then
               updateIfNeeded(SECURITY_SID, "Armed", bitw.band(zoneBypass, 2 ^ (i - 1)) > 0 and 0 or 1, sensor['id'])
            end
         end
      end
   end
   return true
end

-- pmHandlePanel (0xA7)
function pmHandlePanel()
   local i
   local msgCnt = string.byte(pmIncomingPdu, 3)
   for i = 1, msgCnt do
      local eventZone = string.byte(pmIncomingPdu, 3 + 2 * i)
      local logEvent  = string.byte(pmIncomingPdu, 4 + 2 * i)
      local eventType = bitw.band(logEvent, 0x7F)
      local isPanelEvent = (bitw.band(logEvent, 0x80) == 0x80)
      local s = (pmLogEvent_t[pmLang][eventType + 1] or "UNKNOWN") .. " / " .. (pmLogUser_t[eventZone + 1] or "UNKNOWN")
      debug("System message: " .. s)
      local alarmStatus = pmPanelAlarmType_t[eventType] or "None"
      local troubleStatus = pmPanelTroubleType_t[eventType] or "None"
      luup.variable_set(PANEL_SID, "PanelStatusCode", string.format("%02X", logEvent), pmPanelDev)
      luup.variable_set(PANEL_SID, "PanelStatusData", s, pmPanelDev)
      luup.variable_set(PANEL_SID, "PanelAlarmType", alarmStatus, pmPanelDev)
      luup.variable_set(PANEL_SID, "PanelTroubleType", troubleStatus, pmPanelDev)
      --luup.variable_set(PARTITION_SID, "LastUser", pmLogUser_t[eventZone + 1], pmPartitionDev_t[1])
      -- Update siren status
      local noSiren = ((eventType == 0x0B) or (eventType == 0x0C)) and (pmSilentPanic == true)
      if (alarmStatus ~= "None") and (eventType ~= 0x04) and (noSiren == false) then
         for i = 1, #pmSirenDev_t do
            luup.variable_set(SIREN_SID, "Status", "1", pmSirenDev_t[i])
         end
         pmSirenActive = os.time() + 60 * pmBellTime
      end
      if (eventType == 0x1B) and (pmSirenActive ~= nil) then -- Cancel Alarm
         pmSirenActive = nil
         for i = 1, #pmSirenDev_t do
            luup.variable_set(SIREN_SID, "Status", "0", pmSirenDev_t[i])
         end
      end
      if (eventType == 0x60) then -- System reset
         luup.call_delay("pmStartDownload", 2)  -- give PM some time before initiating download
      end
   end
   return true
end

-- pmHandlePowerlink (0xAB)
function pmHandlePowerlink()
   local subType = string.byte(pmIncomingPdu, 3)
   if (subType == 0x03) then -- keepalive message
      pmSendAck(0x02)
      pmLastKeepAlive = os.time()
      pmPowerlinkRetry = 0
   elseif (subType == 0x05) then -- phone message
      local action = string.byte(pmIncomingPdu, 5)
      if (action == 1) then
         pmMessage("Calling user.", 2)
      elseif (action == 2) then
         pmMessage("User acknowledged by phone.", 2)
      end
      pmSendAck(0x02)
   elseif (subType == 0x0A and string.byte(pmIncomingPdu, 5) == 0x01) then
      debug("Enrolling PowerLink")
      pmSendMessage("MSG_ENROLL")
      pmExpectedResponse = string.char(0xAB) .. pmExpectedResponse
   else
      pmSendAck(0x02)
   end
   return true
end

-- RequestArmMode
function RequestArmMode(state, pin)
   local pinPDU, isPL = pmGetPin(pin)
   local armCode = pmArmMode_t[state]
   local allowed = true
   
   debug("RequestArmMode " .. (state or "N/A"))
   if (armCode ~= nil) then
      if (pmRemoteArm == "false") then
         allowed = false
      elseif (pmRemoteArm == "arm") then
         allowed = (bitw.band(armCode, 0x5) == 0x5)
      elseif (pmRemoteArm == "stay") then
         allowed = (bitw.band(armCode, 0x5) ~= 0)
      else
         allowed = true
      end
      if (allowed == true) or (isPL == false) then
         pmSendMessage("MSG_ARM", { arm = string.char(armCode), pin = pinPDU })
      else
         pmMessage((state or "N/A") .. " not allowed without PIN.", 2)
      end
   else
      debug(string.format("RequestArmMode invalid state requested " .. (state or "N/A")))
   end
end

-- RequestQuickArmMode
function RequestQuickArmMode(DetailedArmMode)
	debug("RequestQuickArmMode " .. (state or "N/A"))

   if (DetailedArmMode == "Stay") or (DetailedArmMode == "Armed") then
      if (pmPowerlinkMode == true) and (pmQuickArm == false) then
         pmMessage("Quick arm mode not enabled in panel settings.", 2)
      else
         RequestArmMode(DetailedArmMode, string.char(0, 0))
      end
   else
      pmMessage("RequestQuickArmMode only possible for Arm or Stay.", 2)
	end
end

-- SetArmed
function SetArmed(device, newArmedValue)
   if (pmPowerlinkMode == true) then
      if (pmBypassOff == false) then
         local zone = tonumber(string.sub(luup.devices[device].id, 2))
         local bypass = string.format("%08X", 2 ^ (zone - 1))
         local bypassStr = ""
         for i = 0, 3 do
            bypassStr = bypassStr .. string.char("0x" .. string.sub(bypass, 7 - i * 2, 8 - i * 2))
         end
         if (newArmedValue == "1") then
            pmSendMessage("MSG_BYPASSDIS", { pin = pmPincode_t[1], bypass = bypassStr })
         elseif (newArmedValue == "0") then
            pmSendMessage("MSG_BYPASSEN", { pin = pmPincode_t[1], bypass = bypassStr })
         end
         pmSendMessage("MSG_STATUS") -- request status to check success and update variable
      else
         pmMessage("Bypass option not enabled in panel settings.", 2)
      end
   else
      pmMessage("Bypass setting only supported in Powerlink mode.", 2)
   end
end

-- SetTarget
function SetTarget(device, DeviceID, newTargetValue)
	-- Two ways to call this function
	-- 1. Using lul_device (don't set 'DeviceID')
	-- 2. Setting 'DeviceID' to the X10 nr (e.g. 7 for X07)
   if (DeviceID ~= nil) then
		local s = (DeviceID == 0) and "PGM" or string.format("X%02d", DeviceID)
		local child = findChild(pmPanelDev, s)
		if (child == nil) then
			debug(string.format("Unable to locate device X%02d", DeviceID))
			return false
		else
			device = child
		end
	end
	local device_nr = luup.variable_get(PANEL_SID, "X10DeviceNr", device)
 
	if (device_nr ~= nil) then
	   local onoff = newTargetValue or 0
      local x10dev = string.char(tonumber(string.sub(device_nr, 3, 4),16)) .. string.char(tonumber(string.sub(device_nr, 1, 2),16))
	   debug("Turn X10 device on/off (" .. onoff .. "/" ..  device_nr ..")")
	   luup.variable_set(X10DEVICE_SID, "Target", onoff, device)
      pmSendMessage("MSG_X10PGM", { cmd = string.char(onoff), device = x10dev })
		-- If it went OK, we will receive an A5 status from the PowerMax, which will update the X10 status
	end
end

-- GetTarget
function GetTarget(device, Target)
   luup.variable_set(X10DEVICE_SID, "RetTargetValue", Target, device)
end

-- GetStatus
function GetStatus(device, Status)
   luup.variable_set(X10DEVICE_SID, "ResultStatus", Status, device)
end

-- SetLoadLevelTarget
function SetLoadLevelTarget(device, newLoadlevelTarget)
	local device_nr = luup.variable_get(PANEL_SID, "X10DeviceNr", device)
 
	if (device_nr ~= nil) then
      local current = luup.variable_get(X10DEVICEDIM_SID, "LoadLevelTarget", device) or 0
	   debug("Dimming X10 device (" .. newLoadlevelTarget .. "/" ..  device_nr ..")")

      local cmd = nil
      current = tonumber(current)
      if (newLoadlevelTarget == "0") then
         SetTarget(device, nil, 0)
         luup.variable_set(X10DEVICEDIM_SID, "LoadLevelTarget", "0", device)
         luup.variable_set(X10DEVICEDIM_SID, "LoadLevelStatus", "0", device) -- TODO: only after ACK
      elseif (newLoadlevelTarget == "100") then
         SetTarget(device, nil, 1)
         luup.variable_set(X10DEVICEDIM_SID, "LoadLevelTarget", "100", device)
         luup.variable_set(X10DEVICEDIM_SID, "LoadLevelStatus", "100", device) -- TODO: only after ACK
      elseif (tonumber(newLoadlevelTarget) > current) then
         cmd = 0x0B -- bright
         current = (current < 90) and (current + 10) or 100
      elseif (tonumber(newLoadlevelTarget) < current) then
         cmd = 0x0A -- dim
         current = (current > 10) and (current - 10) or 0
      end
      if (cmd ~= nil) then
         luup.variable_set(X10DEVICEDIM_SID, "LoadLevelTarget", current, device)
         luup.variable_set(X10DEVICEDIM_SID, "LoadLevelStatus", current, device) -- TODO: only after ACK
         local x10dev = string.char(tonumber(string.sub(device_nr, 3, 4),16)) .. string.char(tonumber(string.sub(device_nr, 1, 2),16))
         pmSendMessage("MSG_X10PGM", { cmd = string.char(cmd), device = x10dev })
         pmSendMessage("MSG_X10PGM", { cmd = string.char(cmd), device = x10dev }) -- send it twice (step by 10%)
      end
   end
end

-- GetLoadLevelTarget
function GetLoadLevelTarget(device, Target)
   luup.variable_set(X10DEVICEDIM_SID, "GetLoadlevelTarget", Target, device)
end

-- GetLoadLevelStatus
function GetLoadLevelStatus(device, Status)
   luup.variable_set(X10DEVICEDIM_SID, "retLoadlevelStatus", Status, device)
end

-- ToggleState
function ToggleState(device)
	local lul_arguments = {}
	local onoff = luup.variable_get(X10DEVICE_SID, "Status", device) or 0
	lul_arguments["newTargetValue"] = 1 - onoff
	lul_resultcode, lul_resultstring, lul_job, lul_returnarguments = luup.call_action(X10DEVICE_SID, "SetTarget", lul_arguments, device)
end

-- GetEventLog
function GetEventLog(device, PINCode)
	debug("GetEventLog")
	luup.variable_set(PANEL_SID, "EventLog", "", device)
   pmSendMessage("MSG_EVENTLOG", { pin = pmGetPin(PINCode) })
end
